from pathlib import Path
from typing import Callable, Tuple, Type, Union

import pytest

from config_file import ConfigFile
from config_file.abstract_parser import AbstractParser
from config_file.config_file_path import ConfigFilePath


@pytest.fixture
def tmp_file(tmp_path) -> Callable[[str, str], Path]:
    def func(file_name: str, file_contents: str) -> Path:
        file = tmp_path / file_name
        file.touch()
        file.write_text(file_contents, encoding="utf-8")
        return file

    return func


@pytest.fixture
def template_file(tmp_file: Callable[[str, str], Path]) -> Callable[[str, str], Path]:
    template_dir = Path(__file__).parent / "templates"

    def func(file_type: str, template_name: str = "default") -> Path:
        file_name = f"{template_name}.{file_type}"
        template = template_dir / file_type / file_name

        return tmp_file(file_name, template.read_text(encoding="utf-8"))

    return func


@pytest.fixture
def template_original_file() -> Callable[[Path], Path]:
    def func(template_path: Union[str, Path]):
        if isinstance(template_path, str):
            template_path = Path(template_path)

        original_path = ConfigFilePath(template_path).original_path
        original_path.touch()
        original_path.write_text(template_path.read_text(encoding="utf-8"))

        return original_path

    return func


@pytest.fixture
def template_and_parser(
    template_file: Callable[[str, str], Path]
) -> Callable[[str, str], Tuple[Path, Type[AbstractParser]]]:
    def func(
        file_type: str, template_name: str = "default"
    ) -> Tuple[Path, Type[AbstractParser]]:
        test_file = template_file(file_type, template_name)
        return test_file, ConfigFile(test_file)._ConfigFile__parser

    return func


@pytest.fixture
def templated_parser(template_and_parser) -> Callable[[str, str], Type[AbstractParser]]:
    def func(file_type: str, template_name: str = "default") -> Type[AbstractParser]:
        return template_and_parser(file_type, template_name)[1]

    return func



from types import ModuleType
from typing import Any, Type

from config_file.abstract_parser import AbstractParser
from config_file.exceptions import ParsingError
from config_file.nested_lookup import get_occurrence_of_key
from config_file.utils import split_on_dot


class BaseParser(AbstractParser):
    def __init__(
        self, file_contents: str, module: ModuleType, decode_error: Type[Exception]
    ):
        """The BaseParser implements the AbstractParser for us by giving it any module.

        As long as the module has a `loads` and `dumps` method, the BaseParser can use
        it. Every file is then represented and worked with as a dictionary.

        Args:
            file_contents: The contents of the file to parse.

            module: The module to use to parse the files.
            Must have a `loads` and `dumps` method.

            decode_error: The error the module raises when the
            file cannot be decoded.

        Raises:
            ParsingError: If the decode_error is raised.
        """
        super().__init__(file_contents)
        self.__module = module
        self.__decode_error = decode_error
        self.parsed_content = self.parse_file_contents(module, decode_error)

    def parse_file_contents(self, module: ModuleType, decode_error: Exception) -> dict:
        """Parse the file contents by running the `loads`  method on the module.

        Args:
            module: The module to use to parse the file.
            decode_error: The error that is raised from the module if the file
            cannot be decoded.

        Raises:
            ParsingError: If the decode_error is raised while running `loads`.

        Returns:
            The parsed file as a dictionary.
        """
        try:
            return module.loads(self.file_contents)
        except decode_error as error:
            raise ParsingError(error)

    def reset_internal_contents(self, file_contents: str) -> None:
        """Reset the file contents and parsed contents of the parser.

        Args:
            file_contents: The new file contents.

        Raises:
            ParsingError: If we're unable to parse the new file contents.
        """
        self.file_contents = file_contents
        self.parsed_content = self.parse_file_contents(
            self.__module, self.__decode_error
        )

    def get(self, search_key: str) -> Any:
        key_causing_error = None

        try:
            if "." not in search_key:
                key_causing_error = search_key
                return self.parsed_content[search_key]
            else:
                split_keys = split_on_dot(search_key)

            content_reference = self.parsed_content
            for key in split_keys:
                key_causing_error = key
                content_reference = content_reference[key]

        except (KeyError, TypeError):
            raise KeyError(
                f"cannot `get` {search_key} because "
                f"{key_causing_error} is not subscriptable."
            )

        return content_reference

    def set(self, key: str, value: Any) -> None:
        if "." not in key:
            self.parsed_content[key] = value
            return

        keys = split_on_dot(key)
        content_reference = self.parsed_content
        for index, key in enumerate(keys):
            if index == len(keys) - 1:
                content_reference[key] = value
                return

            try:
                content_reference = content_reference[key]
            except KeyError:
                content_reference[key] = {}
                content_reference = content_reference[key]

    def delete(self, key: str) -> None:
        try:
            if "." not in key:
                del self.parsed_content[key]
            else:
                indexes = split_on_dot(key)
                indexes_length = len(indexes)

                content_reference = self.parsed_content
                for index, key in enumerate(indexes):
                    if index == indexes_length - 1:
                        del content_reference[key]
                        break

                    content_reference = content_reference[key]
        except (KeyError, TypeError):
            raise KeyError(f"The specified key '{key}' to delete was not found.")

    def stringify(self) -> str:
        return self.__module.dumps(self.parsed_content)

    def has(self, search_key: str, wild: bool = False) -> bool:
        if wild:
            return get_occurrence_of_key(self.parsed_content, key=search_key) > 0

        try:
            self.get(search_key)
            return True
        except (KeyError, ParsingError):
            return False



import configparser

from .base_parser import BaseParser
from .ini_parser_compat import IniParserCompatibility


class IniParser(BaseParser):
    def __init__(self, file_contents: str):
        super().__init__(
            file_contents=file_contents,
            module=IniParserCompatibility,
            decode_error=configparser.Error,
        )



from .base_parser import BaseParser


class YamlParser(BaseParser):
    def __init__(self, file_contents: str):
        try:
            from ruamel.yaml import YAML, YAMLError, round_trip_dump, round_trip_load

            yaml = YAML(typ="safe")
            yaml.loads = round_trip_load
            yaml.dumps = round_trip_dump

            super().__init__(
                file_contents=file_contents, module=yaml, decode_error=YAMLError
            )
        except ImportError:
            raise ImportError(
                "It doesn't appear `ruamel.yaml` is installed, but a yaml "
                "file was attempted to be used. Install the `yaml` "
                "extra first with `pip install config-file[yaml]`."
            )



from .base_parser import BaseParser


class TomlParser(BaseParser):
    def __init__(self, file_contents: str):
        try:
            import tomlkit
            from tomlkit.exceptions import ParseError

            super().__init__(
                file_contents=file_contents, module=tomlkit, decode_error=ParseError,
            )
        except ImportError:
            raise ImportError(
                "It doesn't appear `tomlkit` is installed, but a toml "
                "file was attempted to be used. Install the `toml` "
                "extra first with `pip install config-file[toml]`."
            )



import json

from .base_parser import BaseParser


class JsonParser(BaseParser):
    def __init__(self, file_contents: str):
        super().__init__(
            file_contents=file_contents,
            module=json,
            decode_error=json.decoder.JSONDecodeError,
        )

    def stringify(self) -> str:
        return json.dumps(self.parsed_content, indent=4)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['config_file', 'tests']

package_data = \
{'': ['*'],
 'tests': ['templates/ini/*',
           'templates/json/*',
           'templates/toml/*',
           'templates/yaml/*']}

extras_require = \
{'toml': ['tomlkit>=0.6.0,<0.7.0'], 'yaml': ['ruamel.yaml>=0.16.10,<0.17.0']}

setup_kwargs = {
    'name': 'config-file',
    'version': '0.11.0',
    'description': 'Modify configuration files of various formats with the same simple API.',
    'long_description': '# Config File\n\n> Modify configuration files of various formats with the same simple API\n\n![Python Version](https://img.shields.io/pypi/pyversions/config-file.svg)\n[![Version](https://img.shields.io/pypi/v/config-file)](https://pypi.org/project/config-file/)\n[![Style](https://img.shields.io/badge/code%20style-black-000000.svg)](https://pypi.org/project/black/)\n[![Documentation Status](https://readthedocs.org/projects/config-file/badge/?version=latest)](https://config-file.readthedocs.io/en/latest/?badge=latest)\n[![Build Status](https://github.com/eugenetriguba/config-file/workflows/python%20package%20CI/badge.svg?branch=master)](https://github.com/eugenetriguba/config-file/actions/)\n[![Codecov](https://codecov.io/gh/eugenetriguba/config-file/graph/badge.svg)](https://codecov.io/gh/eugenetriguba/config-file)\n\n## About Config File\n\nThe Config File project is designed to allow you to easily manipulate your\nconfiguration files with the same simple API whether they are in INI, JSON,\nYAML, or TOML.\n\n## Installation\n\nConfig File is available to download through PyPI.\n\n```bash\n$ pip install config-file\n```\n\n### Installing Extras\n\nIf you want to manipulate YAML and TOML, you\'ll want to download the extras as well.\n\n```bash\n$ pip install config-file[yaml, toml]\n```\n\nYou can also use [Poetry](https://python-poetry.org).\n\n```bash\n$ poetry install config-file -E yaml -E toml\n```\n\n## Usage\n\nFor this overview, let\'s say you have the following `ini` file\nyou want to manipulate.\n\nDo note, however, that the `ini` format is the oddest format that\n`ConfigFile` supports in that it has no formal specification and is\nnot type aware. When retrieving items from the file, it will return\nthem as strings by default. Others are more type aware and do not\nrequire as much type coercion.\n\n```ini\n[section]\nnum_key = 5\nstr_key = blah\nbool_key = true\nlist_key = [1, 2]\n\n[second_section]\ndict_key = { "another_num": 5 }\n```\n\nIt must have a `.ini` extension in order\nfor the package to recognize it and use the correct parser for it.\n\n### Setting up ConfigFile\n\nTo use the package, we import in the `ConfigFile` object. This object\nis the only thing considered to be part of the public API.\n\nWe can set it up by giving it a string or `pathlib.Path` as the argument.\nAny home tildes `~` in the string or `Path` are recognized and converted\nto the full path for us.\n\n```python\nfrom config_file import ConfigFile\n\nconfig = ConfigFile("~/some-project/config.ini")\n```\n\n### Using `get()`\n\nA recurring pattern you\'ll see here is that all methods that\nneed to specify something inside your configuration file will\ndo so using a dot syntax.\n\n#### Retrieving keys and sections\n\nSo to retrieve our `num_key`, we\'d specify the heading and the\nkey separated by a dot. All values will then be retrieved as\nstrings.\n\n```python\nconfig.get(\'section.num_key\')\n>>> \'5\'\n```\n\nWhile we can retrieves keys, we can also retrieve the entire\nsection, which will be returned back to us as a dictionary.\n\n```python\nconfig.get(\'section\')\n>>> {\'num_key\': \'5\', \'str_key\': \'blah\', \'bool_key\': \'true\', \'list_key\': \'[1, 2]\'}\n```\n\nFurthermore, you can also index into the `ConfigFile` object\nto retrieve keys if that is preferred.\n\n```python\nconfig[\'section\'][\'num_key\']\n>>> \'5\'\n```\n\n#### Coercing the return types\n\nHowever, some of these keys are obviously not strings natively.\nIf we are retrieving a particular value of a key, we may want to\ncoerce it right away without doing clunky type conversions after\neach time we retrieve a value. To do this, we can utilize the\n`return_type` keyword argument.\n\n```python\nconfig.get(\'section.num_key\', return_type=int)\n>>> 5\n```\n\nSometimes we don\'t have structures quite that simple though. What\nif we wanted all the values in `section` coerced? For that, we can\nutilize a `parse_types` keyword argument.\n\n```python\nconfig.get(\'section\', parse_types=True)\n>>> {\'num_key\': 5, \'str_key\': \'blah\', \'bool_key\': True, \'list_key\': [1, 2]}\n```\n\nIt also works for regular keys.\n\n```python\nconfig.get(\'section.num_key\', parse_types=True)\n>>> 5\n```\n\n#### Handling non-existent keys\n\nSometimes we want to retrieve a key but are unsure of if it will exist.\nThere are two ways we could handle that.\n\nThe first is the one we\'re used to seeing: catch the error.\n\n```python\ntry:\n    important_value = config.get(\'section.i_do_not_exist\')\nexcept KeyError:\n    important_value = 42\n```\n\nHowever, the `get` method comes with a `default` keyword argument that we\ncan utilze for this purpose.\n\n```python\nconfig.get(\'section.i_do_not_exist\', default=42)\n>>> 42\n```\n\nThis can be handy if you have a default for a particular configuration value.\n\n### Using `set()`\n\nWe can use `set()` to set a existing key\'s value.\n\n```python\nconfig.set(\'section.num_key\', 6)\n```\n\nThe method does not return anything, since there is nothing\nuseful to return. If something goes wrong where it is unable to set\nthe value, an exception will be raised instead. This is the case\nfor most methods on `ConfigFile`, such as `delete()` or `save()`,\nwhere there would be no useful return value to utilize.\n\nWith `set()`, we can also create and set keys that don\'t exist yet.\n\n```python\nconfig.set(\'new_section.new_key\', \'New key value!\')\n```\n\nWould then result in the following section being added to our original file:\n\n```ini\n[new_section]\nnew_key = New key value!\n```\n\nThe exact behavior of how these new keys or sections are added are a bit\ndependent on the file format we\'re using, since every format is a little\ndifferent in it\'s structure and in what it supports. Mostly though, `ini`\nis just the odd one.\n\nIf we try the following in `ini`, which does not support subsections or\nnested keys, we simply get a single section.\n\n```python\nconfig.set("section.sub_section.sub_sub_section.key", 5)\n```\n\n```ini\n[section.sub_section.sub_sub_section]\nkey = 5\n```\n\nLastly, we can set values using an array notation as well. The underlying\ncontent is all manipulated as a dictionary for every file type. If we wanted\nto create a new section, we\'d simply set it to be an empty dictionary.\n\n```python\nconfig[\'new_section\'] = {}\n```\n\nWhich would result to be an empty section:\n\n```ini\n[new_section]\n```\n\n### Using `delete()`\n\n`delete()` allows us to delete entire sections or specific keys.\n\n```python\nconfig.delete(\'section\')\n```\n\nWould result in the entire section being removed from our configuration file.\nHowever, we can also just delete a single key.\n\n```python\nconfig.delete(\'section.num_key\')\n```\n\nWe can also use the array notation here as well.\n\n```python\ndel config[\'section\'][\'num_key\']\n```\n\n### Using `has()`\n\n`has()` allows us to check whether a given key exists in our file. There\nare two ways to use `has()`.\n\nThe first is using the dot syntax.\n\n```python\nconfig.has(\'section.str_key\')\n>>> True\nconfig.has(\'does_not_exist\')\n>>> False\n```\n\nThis will check if our specific key or section exists. However, we can\nalso check in general if a given key or sections exists anywhere in our\nfile with the `wild` keyword argument.\n\n```python\nconfig.has(\'str_key\', wild=True)\n>>> True\n```\n\n### Using `save()`\n\nFor any changes we make to our configuration file, they are not written out\nto the filesystem until we call `save()`. This is to avoid unnecessary write\ncalls after each operation until we actually need to save.\n\n```python\nconfig.delete(\'section.list_key\')\nconfig.save()\n```\n\n### Using `stringify()`\n\n`stringify()` shows us our configuration file, with any changes we\'ve made,\nas a string. `stringify()` will always show us our latest changes since it\nis stringify-ing our internal representation of the configuration file, not\njust the file we\'ve read in.\n\n```python\nconfig.stringify()\n>>> \'[section]\\nnum_key = 5\\nstr_key = blah\\nbool_key = true\\nlist_key = [1, 2]\\n\\n[second_section]\\ndict_key = { "another_num": 5 }\\n\\n\'\n```\n\n\n### Using `restore_original()`\n\nIf we have a initial configuration file state, we could keep a copy of that\ninitial file and restore back to it whenever needed using `restore_original()`.\n\nBy default, if we created our `ConfigFile` object with the path of `~/some-project/config.ini`,\n`restore_original()` will look for our original file at `~/some-project/config.original.ini`.\n\n```python\nconfig.restore_original()\n```\n\nHowever, if we have a specific path elsewhere that this original configuration file is or it\nis named differently than what the default expects, we can utilize the `original_path`\nkeyword argument.\n\n```python\nconfig.restore_original(original_path="~/some-project/original-configs/config.ini")\n```\n\n\n## Format Versions Supported\n\n| Format  | Specification version supported |\n| ------------- | ------------- |\n| INI  | No official specification. |\n| JSON  | [RFC 7159](https://tools.ietf.org/html/rfc7159.html)  |\n| YAML  | [v1.2](https://yaml.org/spec/1.2/spec.html)  |\n| TOML  | [v1.0.0-rc.1](https://github.com/toml-lang/toml/releases/tag/v1.0.0-rc.1)  |\n\nFor `ini` and `json`, Python\'s standard library modules are used.\nRegarding `ini`, there is no formal specification so the syntax that configparser\nsupports is what is supported here.\n\n## License\n\nThe [MIT](https://github.com/eugenetriguba/config-file/blob/master/LICENSE) License.\n',
    'author': 'Eugene Triguba',
    'author_email': 'eugenetriguba@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/eugenetriguba/config_file',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import configparser
from io import StringIO


class IniParserCompatibility:
    """
    Provides compatibility with the json, yaml,
    and toml API so we can use the BaseParser
    easily, while still being able to leverage
    the built-in configparser to parse the file.

    It is used as the `module` parameter
    for BaseParser, which requires us to have a
    `dumps` and `loads` method.
    """

    @staticmethod
    def loads(file_contents: str) -> dict:
        parser = configparser.ConfigParser()
        parser.read_string(file_contents)

        return IniParserCompatibility.__create_configparser_dict(parser)

    @staticmethod
    def dumps(file_contents: dict) -> str:
        parser = configparser.ConfigParser()
        buffer = StringIO()

        parser.read_dict(file_contents)
        parser.write(buffer)

        return buffer.getvalue()

    @staticmethod
    def __create_configparser_dict(parser: configparser.ConfigParser) -> dict:
        result = {}
        items = dict(parser.items())

        for section in parser.sections():
            result.update({str(section): dict(items[section])})

        return result
